module PowerMod(powermod) where

import Data.Bits

{- | @powermod m b e@ = b^e mod m &#8805; 0

   Assumptions: m>0, b unrestricted, e&#8805;0.

   The argument order is such that you usually benefit from:

   > let (^^) = powermod 101 in (2^^20, 3^^30, 4^^40)
-}
powermod :: (Bits i) => Integer -> Integer -> i -> Integer
powermod m b 0 = 1
powermod m b e = let b' = b `mod` m in f b' (e-1) b'
  where f _ 0 y = y
        f x n y = g x n
          where g x n | not (n `testBit` 0) = g (x*x `rem` m) (n `shiftR` 1)
                      | otherwise = f x (n `clearBit` 0) (x*y `rem` m)
{-
This algorithm mirrors the x^n algorithm in the standard Haskell prelude.
To see how it works, here are some lemmata:
f x n y = x^n * y
If n>0, g x n = x^n * y

The beginning uses `mod` to ensure non-negativeness.
Afterwards `rem` because it is rumoured to be faster.
-}
