{- | This module produces tables of small primes.
-}
module SmallPrimes(smallprimes) where

import Control.Monad(when)
import Control.Monad.ST
import Data.Array.ST
import Data.Array.Unboxed

prime_a n = do s <- newArray (2,n) True
               sieve s 2
               return s
    where rn = floor (sqrt (fromIntegral n))
          sieve s m | m > rn = return ()
                    | otherwise = do b <- readArray s m
                                     when b (mapM_ (\i -> writeArray s i False) [2*m,3*m..n])
                                     sieve s (m+1)

{- | @smallprimes n@ gives:

   * list of primes in 2..n, in increasing order, in both Int and Integer types.

   * array with domain 2..n, and position i True iff i is prime
-}
smallprimes :: Int -> ([Int], [Integer], UArray Int Bool)
smallprimes n = runST p
  where p = do a <- prime_a n :: ST s (STUArray s Int Bool)
               b <- unsafeFreeze a
               let s = [i | (i,True) <- assocs b]
               return (s, map toInteger s, b)

