{- | This module provides the test for strong probable primes,
   i.e., the Rabin-Miller test.

   Let n be given, odd, and n-1 = k*2^a, k odd.
   n is a strong probable prime of base b iff

   * b^k mod n = 1, or

   * for some i in 0..a-1, b^(k*2^i) mod n = n-1
-}
module StrongProbablePrime(spp) where

import TrialDivision(efactor)
import PowerMod

{- | @spp n b@ determines whether n is a strong probable prime of base b.
-}
spp :: Integer -> Integer -> Bool
spp n b = bk == 1 || any (== n-1) s
  where (2,a):ks = efactor [2] (n-1)
        k = case ks of { [] -> 1; [(k,1)] -> k }
        bk = powermod n b k
        s = take (fromInteger a) (iterate (\x -> (x*x) `rem` n) bk)
