{- Looks for a large prime.
-}
module Main(main) where

import Control.Monad(forM_, msum)
import Control.Monad.ST
import Data.Array.ST
import Data.List(find)
import Data.Maybe(fromJust)
import System.Environment(getArgs, getProgName)
import System.IO(hFlush, hPutStr, hPutStrLn, stdout, stderr)
import System.Random(randomRIO)

import SmallPrimes
import StrongProbablePrime

-- we have all small primes below this
primeBaseBound = 500000
-- strong probable prime test tries this many small primes
probPrimeStrength = 2000

main = do
  as <- getArgs
  case as of
    [a0] -> do
      let n0 = read a0
          r0 = 5 * length (show n0)   -- see the end for this estimate
          p = (fromJust . msum .
               map (flip firstPrime r0) .
               iterate (+ fromIntegral r0)) n0
      print p
    _ -> do p <- getProgName
            putStrLn (p ++ " start")
            putStrLn "Looks for the first prime after start"

ePutStr s = hPutStr stderr s
ePutStrLn s = hPutStrLn stderr s

primeBase :: [Integer]
(_,primeBase,_) = smallprimes primeBaseBound

probPrime :: Integer -> Bool
probPrime n = all (spp n) (take probPrimeStrength primeBase)

firstPrime :: Integer -> Int -> Maybe Integer
firstPrime n r | n <= last primeBase = find (n <=) primeBase
               | otherwise = runST (primeSearch n r)

primeSearch n r = do
  s <- newArray (0,r-1) True
  forM_ primeBase (mark s n r)
  ps <- (map ((n + ) . toInteger . fst) . filter snd) `fmap` getAssocs s
  return (find probPrime ps)

mark :: STUArray s Int Bool -> Integer -> Int -> Integer -> ST s ()
mark s n r p = do
  let n' = fromInteger (n `rem` p)
      a = if n'==0 then n' else p' - n'
      p' = fromInteger p
  forM_ [a, a+p' .. r-1] (\i -> writeArray s i False)


{- A note is in order for the simple estimate of r0, the sieving range.

First we crudely assume the number of primes below x is x/ln x. Yes
yes, it's relative. We can always make up for it by fudge factors and
unbounded search elsewhere.

Then we solve for r0 in

  (n0+r0) / ln(n0+r0) - n0 / ln n0 >= 1

Using the linear approximation

  ln(n0+r0) = ln(1 + r0/n0) + ln n0 = r0/n0 + error + ln n0

we get

  r0 >= ln n0 / (1 - 1/ln n0 - 1/n0)

We further approximate that as ln n0: the 1/n0 there does nothing when
we get to like 10-digit numbers, and the 1/ln n0 there only contributes
a +/-1 (recall series expansion of y/(1-y)) --- easily dwarved by our
very crude premise above and fudge factor below.

Lastly,

  ln n0 = log_10 n0 * ln 10
        = (length of n0 +/- 1) * 2.3...

and let's throw in a more-than-double fudge multiplier and just call it
length of n0 * 5.

This interval width is almost always more than enough. In the rare
case it isn't, note that in the rest of the program we keep going for
the next interval, then the interval after next, ... It won't take too
many intervals even in the worse case.

-}
